/*
 * Copyright (C) 2012 National Instruments
 * All rights reserved.
 * Author: Sankara Muthukrishnan <sankara.m@gmail.com>
 *
 * Tests the timerfd feature and measures min,max and average cost of 
 * timerfd_settime and "read API in blocking mode"
 *
 * uses some functions/code cyclictest (in rt_tests suite) that has
 * the following copyright notices:
 *
 * (C) 2008-2011 Clark Williams <williams@redhat.com>
 * (C) 2005-2007 Thomas Gleixner <tglx@linutronix.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License Version
 * 2 as published by the Free Software Foundation.
 * 
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <malloc.h>
#include <pthread.h>
#include <dlfcn.h>
#include <unistd.h>
#include <sys/syscall.h>
#include <stdint.h>
#include <getopt.h>
#include <stdlib.h>
#include <stdarg.h>
#include <limits.h>
#include <sys/timerfd.h>
#include <ctype.h>

#if DEBUG
#define DEBUGA(x) x
#else
#define DEBUGA(x) 
#endif

#define USEC_PER_SEC    1000000
#define NSEC_PER_SEC    1000000000
#define HISTOGRAM_SIZE  50

#define PRINTL printf("%s:%d\n", __FUNCTION__, __LINE__)
#define PRINTFL printf("%s:%d\n", __FILE__, __LINE__)

struct tCmdLineArg{
	int nTestIters;
	int prio;
	int timeout_us;
	int affinity;
	int sched_policy;
	int debug;
	int use_absolute;
};

tCmdLineArg sg_cmdArg;

void warn(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	fputs("WARNING: ", stderr);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
}

void fatal(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	fputs("FATAL: ", stderr);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	exit(EXIT_FAILURE);
}


struct tStat{
	int64_t min;
	int64_t min_iter;
	int64_t max;
	int64_t max_iter;
	double avg;
	int64_t hist_size;
	int64_t *hist;
};

struct thread_param {
	void *event;
	pthread_mutex_t mutex;
	int64_t timeout_ns;
	int cpu;
	int sched_policy;
	int timerfd;
	int prio;
	int count; // number of test iterations
	tStat stat; 
	int debug;
	int use_absolute; // currently there is no command-line argument, but both absolute
	                  // and relative hang the same way 
};

static inline int64_t calcdiff_ns(struct timespec *t1, struct timespec *t2)
{
	int64_t diff;
	diff = NSEC_PER_SEC * (int64_t)((int) t2->tv_sec - (int) t1->tv_sec);
	diff += ((int) t2->tv_nsec - (int) t1->tv_nsec);
	return diff;
}

int set_sched_params_thread(int prio, int policy, int cpu)
{
	struct sched_param schedp;
	cpu_set_t mask;
	
	if (policy != SCHED_OTHER) {
		memset(&schedp, 0, sizeof(schedp));
		schedp.sched_priority = prio;
		if (0 != sched_setscheduler(0, policy, &schedp))
			fatal("Could not set the scheduler policy and priority");
	}

	if (cpu != -1) {
		CPU_ZERO(&mask);
		CPU_SET(cpu, &mask);
		if(sched_setaffinity(0, sizeof(mask), &mask) == -1)
			fatal("Could not set CPU affinity to CPU #%d\n", cpu);
	}
	return 0;
}

void init_thread_param(struct thread_param *par)
{
	par->stat.min = INT_MAX;
	par->stat.max = 0;
	par->stat.avg = 0.0;
	par->stat.hist_size = HISTOGRAM_SIZE;
	par->stat.hist = (int64_t *)calloc(par->stat.hist_size, sizeof(*par->stat.hist));
	par->count = sg_cmdArg.nTestIters;
	par->prio = sg_cmdArg.prio;
	par->timeout_ns = sg_cmdArg.timeout_us * 1000;
	par->cpu = sg_cmdArg.affinity;
	par->sched_policy = sg_cmdArg.sched_policy;
	par->debug = sg_cmdArg.debug;
}

void cleanup_thread_param(struct thread_param *par)
{
	free(par->stat.hist);
}

void calc_stat(int64_t diff_ns, struct thread_param *par, int64_t iter)
{
	if (diff_ns < par->stat.min) {
		par->stat.min = diff_ns;
		par->stat.min_iter = iter;
	}
	if (diff_ns > par->stat.max) {
		par->stat.max = diff_ns;
		par->stat.max_iter = iter;
	}
	par->stat.avg += (double) diff_ns;
	int64_t diff_us = diff_ns / 1000;
	if (diff_us >= (par->stat.hist_size - 1))
		par->stat.hist[par->stat.hist_size-1]++;
	else
		par->stat.hist[diff_us]++;
}

void print_stat(struct thread_param *par, const char *str)
{
	printf("%s. min=%lld us(iter:%lld), max=%lld us(iter:%lld), avg=%lld us\n", str,
			par->stat.min/1000, par->stat.min_iter, 
			par->stat.max/1000, par->stat.max_iter,
			(int64_t) par->stat.avg/1000);
	for (int i = 0; i < par->stat.hist_size; i++) {
		printf("%3d ", (int32_t)par->stat.hist[i]);
		if (!((i+1)%10))
			printf("\n");
	}
	printf("\n");
}

static inline void add_time(struct timespec *tdest, struct timespec *tsrc, int64_t interval)
{
	tdest->tv_nsec = (long) ((uint64_t)tsrc->tv_nsec + (uint64_t)interval) % NSEC_PER_SEC;
	tdest->tv_sec = tsrc->tv_sec + (time_t)(((uint64_t)tsrc->tv_nsec + (uint64_t)interval) / NSEC_PER_SEC);
}

static void *tfd_thread(void *param)
{
	struct thread_param *par = (struct thread_param*) param;
	struct timespec t1, t2;

	set_sched_params_thread(par->prio, par->sched_policy, par->cpu);
	sched_yield();

	struct itimerspec ts;
	ts.it_interval.tv_sec = 0;
	ts.it_interval.tv_nsec = 0;
	ts.it_value.tv_sec = 0;
	ts.it_value.tv_nsec = par->timeout_ns;

	for(int i = 0; i < par->count; i++) {
		clock_gettime(CLOCK_MONOTONIC, &t1);
		if (par->debug)
			printf("Before timerfd_settime\n");
		if (par->use_absolute) {
			add_time(&ts.it_value, &t1, par->timeout_ns);
			if (0 != timerfd_settime(par->timerfd, TFD_TIMER_ABSTIME, &ts, NULL))
				fatal("timerfd_settime failed");
		}
		else {
			if (0 != timerfd_settime(par->timerfd, 0, &ts, NULL))
				fatal("timerfd_settime failed");
		}
		if (par->debug)
			printf("After timerfd_settime\n");
		uint64_t numexp = 0; // number of expirations
		ssize_t s;
		if (par->debug)
			printf("Before read\n");
		s = read(par->timerfd, &numexp, sizeof(numexp));
		if (par->debug)
			printf("after read. numexp=%llu\n", numexp);
		if (s != sizeof(numexp) || numexp == 0)
			fatal("read on timerfd failed, s=%d, numexp=%llu, errno=%d\n", s, numexp, errno);
		clock_gettime(CLOCK_MONOTONIC, &t2); 
		int64_t diff_ns = calcdiff_ns(&t1, &t2);
		diff_ns -= par->timeout_ns;
		calc_stat(diff_ns, par, i);
	}
	par->stat.avg /= par->count;	
	print_stat(par, "wait_thread");
	return 0;
}

int test_timerfd()
{
	int status;
	pthread_t tfd_thr;
	struct thread_param *tfd_par = NULL;

	tfd_par = (struct thread_param *) calloc(1, sizeof(struct thread_param));
	init_thread_param(tfd_par);

	if (tfd_par->debug)
		printf("Before timerfd_create\n");
   if ((tfd_par->timerfd= timerfd_create(CLOCK_MONOTONIC, 0)) == -1)
		fatal("timerfd_create failed");
	if (tfd_par->debug)
		printf("After timerfd_create.Success. fd=%d\n", tfd_par->timerfd);
	status = pthread_create(&tfd_thr, NULL, tfd_thread, tfd_par); 
	pthread_join(tfd_thr, NULL); 
	cleanup_thread_param(tfd_par);
	free(tfd_par);
	close(tfd_par->timerfd);
	return 0;
}

static void display_help(int error)
{
	printf("Usage:\n"
	       "testTimerfd <options>\n\n"
	       "-a [NUM] --affinity           run thread on processor #N\n"
	       "                              if not specified, no cpu affinity is set\n"
	       "-d        --debug             print some debug info about API success, etc - useful for debugging\n"
	       "-n NUM    --num_iters=NUM     run the test for NUM iterations\n"
	       "-p PRIO   --priority=PRIO     priority of the thread (used only for SCHED_FIFO and SCHED_RR policies\n"
	       "                              and ignored for SCHED_OTHER policy)\n"   
	       "-s POLICY --scheduler=POLICY  use the scheduler policy\n"
	       "                              O = SCHED_OTHER\n"
	       "                              F = SCHED_FIFO\n"
	       "                              R = SCHED_RR\n"
	       "-t USEC  --timeout=USEC       set the timer to expire for USEC in the call to timerfd_setttime\n"
	       "-?       --help               display this help\n");
	exit(error);
}

static void process_options (int argc, char *argv[])
{
	int max_cpus = sysconf(_SC_NPROCESSORS_CONF);
	int affinity = -1;
	char sched = 'F';

	sg_cmdArg.debug = 0;
	sg_cmdArg.use_absolute = 0;
	sg_cmdArg.prio = 1;
	sg_cmdArg.nTestIters = 5;
	for (;;) {
		int option_index = 0;
		/** Options for getopt */
		static struct option long_options[] = {
			{"affinity", required_argument, NULL, 'a'},
			{"debug", no_argument, NULL, 'd'},
			{"num_iters", required_argument, NULL, 'n'},
			{"priority", required_argument, NULL, 'p'},
			{"scheduler", required_argument, NULL, 's'},
			{"timeout", required_argument, NULL, 't'},
			{"help", no_argument, NULL, '?'},
			{NULL, 0, NULL, 0}
		}; 

		int c = getopt_long(argc, argv, "a:dn:p:s:t:", long_options, &option_index);
		if (c == -1)
			break;
		switch (c) {
		case 'n':
			sg_cmdArg.nTestIters = atoi(optarg);
			break;
		case 'p':
			sg_cmdArg.prio = atoi(optarg);
			break;
		case 't':
			sg_cmdArg.timeout_us = atoi(optarg);
			break;
		case 'a':
			affinity = atoi(optarg);
			if (affinity < 0 || affinity >= max_cpus) {
				display_help(0); 
				break;
			}
			break;
		case 's':
			sched = toupper(optarg[0]);
			break;
		case 'd':
			sg_cmdArg.debug = 1;
			break;
		case '?': display_help(0); break;
		}
	}
	sg_cmdArg.affinity = affinity;
	switch (sched) {
	case 'F' : 
		sg_cmdArg.sched_policy = SCHED_FIFO;
		break;
	case 'R':
		sg_cmdArg.sched_policy = SCHED_RR;
		break;
	case 'O':
		sg_cmdArg.sched_policy = SCHED_OTHER;
		break;
	default:
		display_help(0); 
		break;
	}
}

void print_cmdline_args()
{
	printf("nTestIters=%d\n", sg_cmdArg.nTestIters);
	printf("prio=%d\n", sg_cmdArg.prio);
	printf("timeout_us=%d\n", sg_cmdArg.timeout_us);
	printf("affinity=%d\n", sg_cmdArg.affinity);
	switch(sg_cmdArg.sched_policy) {
		case SCHED_FIFO:
			printf("scheduler policy=SCHED_FIFO\n");
			break;
		case SCHED_RR:
			printf("scheduler policy=SCHED_RR\n");
			break;
		case SCHED_OTHER:
			printf("scheduler policy=SCHED_OTHER\n");
			break;
	}
	printf("use_absolute(in timerfd_settime)=%d\n", sg_cmdArg.use_absolute);
}

int main(int argc, char *argv[])
{
	process_options(argc, argv);
	print_cmdline_args();
	test_timerfd();
	return 0;
}
